from __future__ import annotations

__all__ = ["JobDB"]

from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, Iterable

from sqlalchemy import bindparam, case, delete, insert, literal, select, update

if TYPE_CHECKING:
    from sqlalchemy.sql.elements import BindParameter
from sqlalchemy.sql import expression

from diracx.core.exceptions import InvalidQueryError
from diracx.core.models.job import JobCommand
from diracx.core.models.search import SearchSpec, SortSpec

from ..utils import BaseSQLDB, _get_columns
from ..utils.functions import utcnow
from .schema import (
    HeartBeatLoggingInfo,
    InputData,
    JobCommands,
    JobDBBase,
    JobJDLs,
    Jobs,
)


class JobDB(BaseSQLDB):
    metadata = JobDBBase.metadata

    # Field names which should be stored in the HeartBeatLoggingInfo table
    heartbeat_fields = {
        "LoadAverage",
        "MemoryUsed",
        "Vsize",
        "AvailableDiskSpace",
        "CPUConsumed",
        "WallClockTime",
    }

    # TODO: this is copied from the DIRAC JobDB
    # but is overwritten in LHCbDIRAC, so we need
    # to find a way to make it dynamic
    jdl_2_db_parameters = ["JobName", "JobType", "JobGroup"]

    async def summary(
        self, group_by: list[str], search: list[SearchSpec]
    ) -> list[dict[str, str | int]]:
        """Get a summary of the jobs."""
        return await self._summary(table=Jobs, group_by=group_by, search=search)

    async def search(
        self,
        parameters: list[str] | None,
        search: list[SearchSpec],
        sorts: list[SortSpec],
        *,
        distinct: bool = False,
        per_page: int = 100,
        page: int | None = None,
    ) -> tuple[int, list[dict[Any, Any]]]:
        """Search for jobs in the database."""
        return await self._search(
            table=Jobs,
            parameters=parameters,
            search=search,
            sorts=sorts,
            distinct=distinct,
            per_page=per_page,
            page=page,
        )

    async def create_job(self, compressed_original_jdl: str):
        """Insert a new job with original JDL. Returns inserted job id."""
        result = await self.conn.execute(
            insert(JobJDLs).values(
                JDL="",
                JobRequirements="",
                OriginalJDL=compressed_original_jdl,
            )
        )
        return result.lastrowid

    async def delete_jobs(self, job_ids: list[int]):
        """Delete jobs from the database."""
        stmt = delete(JobJDLs).where(JobJDLs.job_id.in_(job_ids))
        await self.conn.execute(stmt)

    async def insert_input_data(self, lfns: dict[int, list[str]]):
        """Insert input data for jobs."""
        await self.conn.execute(
            insert(InputData),
            [
                {
                    "JobID": job_id,
                    "LFN": lfn,
                }
                for job_id, lfns_ in lfns.items()
                for lfn in lfns_
            ],
        )

    async def insert_job_attributes(self, jobs_to_update: dict[int, dict]):
        """Insert the job attributes."""
        await self.conn.execute(
            insert(Jobs),
            [
                {
                    "JobID": job_id,
                    **attrs,
                }
                for job_id, attrs in jobs_to_update.items()
            ],
        )

    async def update_job_jdls(self, jdls_to_update: dict[int, str]):
        """Update the JDL, typically just after inserting the original JDL, or rescheduling, for example."""
        await self.conn.execute(
            update(JobJDLs).where(JobJDLs.__table__.c.JobID == bindparam("b_JobID")),
            [
                {
                    "b_JobID": job_id,
                    "JDL": compressed_jdl,
                }
                for job_id, compressed_jdl in jdls_to_update.items()
            ],
        )

    async def set_job_attributes(self, job_data):
        """Update the parameters of the given jobs."""
        # TODO: add myDate and force parameters.

        if not job_data:
            # nothing to do!
            raise ValueError("job_data is empty")

        for job_id in job_data.keys():
            if "Status" in job_data[job_id]:
                job_data[job_id].update(
                    {"LastUpdateTime": datetime.now(tz=timezone.utc)}
                )
        columns = set(key for attrs in job_data.values() for key in attrs.keys())
        case_expressions = {
            column: case(
                *[
                    (
                        Jobs.__table__.c.JobID == job_id,
                        # Since the setting of the new column value is obscured by the CASE statement,
                        # ensure that SQLAlchemy renders the new column value with the correct type
                        literal(attrs[column], type_=Jobs.__table__.c[column].type)
                        if not isinstance(attrs[column], expression.FunctionElement)
                        else attrs[column],
                    )
                    for job_id, attrs in job_data.items()
                    if column in attrs
                ],
                else_=getattr(Jobs.__table__.c, column),  # Retain original value
            )
            for column in columns
        }

        stmt = (
            Jobs.__table__.update()
            .values(**case_expressions)
            .where(Jobs.__table__.c.JobID.in_(job_data.keys()))
        )
        await self.conn.execute(stmt)

    async def get_job_jdls(self, job_ids, original: bool = False) -> dict[int, str]:
        """Get the JDLs for the given jobs."""
        if original:
            stmt = select(JobJDLs.job_id, JobJDLs.original_jdl).where(
                JobJDLs.job_id.in_(job_ids)
            )
        else:
            stmt = select(JobJDLs.job_id, JobJDLs.jdl).where(
                JobJDLs.job_id.in_(job_ids)
            )

        return {jobid: jdl for jobid, jdl in (await self.conn.execute(stmt)) if jdl}

    async def set_job_commands(self, commands: list[tuple[int, str, str]]) -> None:
        """Store a command to be passed to the job together with the next heart beat."""
        await self.conn.execute(
            insert(JobCommands),
            [
                {
                    "JobID": job_id,
                    "Command": command,
                    "Arguments": arguments,
                    "ReceptionTime": datetime.now(tz=timezone.utc),
                }
                for job_id, command, arguments in commands
            ],
        )

    async def set_properties(
        self, properties: dict[int, dict[str, Any]], update_timestamp: bool = False
    ) -> int:
        """Update the job parameters.

        All the jobs must update the same properties.

        :param properties: {job_id : {prop1: val1, prop2:val2}
        :param update_timestamp: if True, update the LastUpdate to now

        :return rowcount

        """
        # Check that all we always update the same set of properties
        required_parameters_set = {tuple(sorted(k.keys())) for k in properties.values()}

        if len(required_parameters_set) != 1:
            raise NotImplementedError(
                "All the jobs should update the same set of properties"
            )

        required_parameters = list(required_parameters_set)[0]
        update_parameters = [{"job_id": k, **v} for k, v in properties.items()]

        columns = _get_columns(Jobs.__table__, required_parameters)
        values: dict[str, BindParameter[Any] | datetime] = {
            c.name: bindparam(c.name) for c in columns
        }
        if update_timestamp:
            values["LastUpdateTime"] = datetime.now(tz=timezone.utc)

        stmt = update(Jobs).where(Jobs.job_id == bindparam("job_id")).values(**values)
        rows = await self.conn.execute(stmt, update_parameters)

        return rows.rowcount

    async def add_heartbeat_data(
        self, job_id: int, dynamic_data: dict[str, str]
    ) -> None:
        """Add the job's heartbeat data to the database.

        NOTE: This does not update the HeartBeatTime column in the Jobs table.
        This is instead handled by the `diracx.logic.jobs.status.set_job_statuses`
        as it involves updating multiple databases.

        :param job_id: the job id
        :param dynamic_data: mapping of the dynamic data to store,
            e.g. {"AvailableDiskSpace": 123}
        """
        if extra_fields := set(dynamic_data) - self.heartbeat_fields:
            raise InvalidQueryError(
                f"Not allowed to store heartbeat data for: {extra_fields}. "
                f"Allowed keys are: {self.heartbeat_fields}"
            )
        values = [
            {
                "JobID": job_id,
                "Name": key,
                "Value": value,
                "HeartBeatTime": utcnow(),
            }
            for key, value in dynamic_data.items()
        ]
        await self.conn.execute(insert(HeartBeatLoggingInfo).values(values))

    async def get_job_commands(self, job_ids: Iterable[int]) -> list[JobCommand]:
        """Get a command to be passed to the job together with the next heartbeat.

        :param job_ids: the job ids
        :return: mapping of job id to list of commands
        """
        # Get the commands
        stmt = (
            select(JobCommands.job_id, JobCommands.command, JobCommands.arguments)
            .where(JobCommands.job_id.in_(job_ids), JobCommands.status == "Received")
            .order_by(JobCommands.job_id)
        )
        commands = await self.conn.execute(stmt)
        # Update the status of the commands
        stmt = (
            update(JobCommands)
            .where(JobCommands.job_id.in_(job_ids))
            .values(Status="Sent")
        )
        await self.conn.execute(stmt)
        # Return the commands grouped by job id
        return [
            JobCommand(job_id=cmd.JobID, command=cmd.Command, arguments=cmd.Arguments)
            for cmd in commands
        ]



from __future__ import annotations

import hashlib
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING

from sqlalchemy import DateTime
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import expression

if TYPE_CHECKING:
    from sqlalchemy.types import TypeEngine


class utcnow(expression.FunctionElement):  # noqa: N801
    type: TypeEngine = DateTime()
    inherit_cache: bool = True


@compiles(utcnow, "postgresql")
def pg_utcnow(element, compiler, **kw) -> str:
    return "TIMEZONE('utc', CURRENT_TIMESTAMP)"


@compiles(utcnow, "mssql")
def ms_utcnow(element, compiler, **kw) -> str:
    return "GETUTCDATE()"


@compiles(utcnow, "mysql")
def mysql_utcnow(element, compiler, **kw) -> str:
    return "(UTC_TIMESTAMP)"


@compiles(utcnow, "sqlite")
def sqlite_utcnow(element, compiler, **kw) -> str:
    return "DATETIME('now')"


class days_since(expression.FunctionElement):  # noqa: N801
    """Sqlalchemy function to get the number of days since a given date.

    Primarily used to be able to query for a specific resolution of a date e.g.

        select * from table where days_since(date_column) = 0
        select * from table where days_since(date_column) = 1
    """

    type = DateTime()
    inherit_cache = False

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)


@compiles(days_since, "postgresql")
def pg_days_since(element, compiler, **kw):
    return f"EXTRACT(DAY FROM (now() - {compiler.process(element.clauses)}))"


@compiles(days_since, "mysql")
def mysql_days_since(element, compiler, **kw):
    return f"DATEDIFF(NOW(), {compiler.process(element.clauses)})"


@compiles(days_since, "sqlite")
def sqlite_days_since(element, compiler, **kw):
    return f"julianday('now') - julianday({compiler.process(element.clauses)})"


def substract_date(**kwargs: float) -> datetime:
    return datetime.now(tz=timezone.utc) - timedelta(**kwargs)


def hash(code: str):
    return hashlib.sha256(code.encode()).hexdigest()



from __future__ import annotations

from datetime import datetime
from typing import Optional

from sqlalchemy import (
    Double,
    Index,
    String,
    Text,
)
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from diracx.db.sql.utils import (
    EnumBackedBool,
    str32,
    str128,
    str255,
)
from diracx.db.sql.utils.types import SmarterDateTime


class PilotAgentsDBBase(DeclarativeBase):
    type_annotation_map = {
        str32: String(32),
        str128: String(128),
        str255: String(255),
    }


class PilotAgents(PilotAgentsDBBase):
    __tablename__ = "PilotAgents"

    pilot_id: Mapped[int] = mapped_column(
        "PilotID", autoincrement=True, primary_key=True
    )
    initial_job_id: Mapped[int] = mapped_column("InitialJobID", default=0)
    current_job_id: Mapped[int] = mapped_column("CurrentJobID", default=0)
    pilot_job_reference: Mapped[str255] = mapped_column(
        "PilotJobReference", default="Unknown"
    )
    pilot_stamp: Mapped[str32] = mapped_column("PilotStamp", default="")
    destination_site: Mapped[str128] = mapped_column(
        "DestinationSite", default="NotAssigned"
    )
    queue: Mapped[str128] = mapped_column("Queue", default="Unknown")
    grid_site: Mapped[str128] = mapped_column("GridSite", default="Unknown")
    vo: Mapped[str128] = mapped_column("VO")
    grid_type: Mapped[str32] = mapped_column("GridType", default="LCG")
    benchmark: Mapped[float] = mapped_column("BenchMark", Double, default=0.0)
    submission_time: Mapped[Optional[datetime]] = mapped_column(
        "SubmissionTime", SmarterDateTime
    )
    last_update_time: Mapped[Optional[datetime]] = mapped_column(
        "LastUpdateTime", SmarterDateTime
    )
    status: Mapped[str32] = mapped_column("Status", default="Unknown")
    status_reason: Mapped[str255] = mapped_column("StatusReason", default="Unknown")
    accounting_sent: Mapped[bool] = mapped_column(
        "AccountingSent", EnumBackedBool(), default=False
    )

    __table_args__ = (
        Index("PilotJobReference", "PilotJobReference"),
        Index("Status", "Status"),
        Index("Statuskey", "GridSite", "DestinationSite", "Status"),
    )


class JobToPilotMapping(PilotAgentsDBBase):
    __tablename__ = "JobToPilotMapping"

    pilot_id: Mapped[int] = mapped_column("PilotID", primary_key=True)
    job_id: Mapped[int] = mapped_column("JobID", primary_key=True)
    start_time: Mapped[datetime] = mapped_column("StartTime", SmarterDateTime)

    __table_args__ = (Index("JobID", "JobID"), Index("PilotID", "PilotID"))


class PilotOutput(PilotAgentsDBBase):
    __tablename__ = "PilotOutput"

    pilot_id: Mapped[int] = mapped_column("PilotID", primary_key=True)
    std_output: Mapped[str] = mapped_column("StdOutput", Text)
    std_error: Mapped[str] = mapped_column("StdError", Text)



from __future__ import annotations

import pytest
from pytest_lazy_fixtures import lf

from diracx.core.exceptions import InvalidQueryError
from diracx.testing.mock_osdb import MockOSDBMixin
from diracx.testing.osdb import DummyOSDB


@pytest.fixture(
    params=[
        pytest.param(lf("prefilled_dummy_opensearch_db"), id="dummy_opensearch_db"),
        pytest.param(lf("prefilled_sql_opensearch_db"), id="sql_opensearch_db"),
    ]
)
def prefilled_db(request):
    """Provide a prefilled database for testing."""
    return request.param


async def test_specified_parameters(prefilled_db: DummyOSDB):
    doc1, doc2, doc3 = prefilled_db.test_docs

    results = await prefilled_db.search(None, [], [])
    assert len(results) == 3
    assert doc1 in results and doc2 in results and doc3 in results

    results = await prefilled_db.search([], [], [])
    assert len(results) == 3
    assert doc1 in results and doc2 in results and doc3 in results

    results = await prefilled_db.search(["IntField"], [], [])
    expected_results = []
    for doc in [doc1, doc2, doc3]:
        expected_doc = {key: doc[key] for key in {"IntField"}}
        # Ensure the document is not already in the list
        # If it is the all() check below no longer makes sense
        assert expected_doc not in expected_results
        expected_results.append(expected_doc)
    assert len(results) == len(expected_results)
    assert all(result in expected_results for result in results)

    results = await prefilled_db.search(["IntField", "UnknownField"], [], [])
    expected_results = [
        {"IntField": doc1["IntField"], "UnknownField": doc1["UnknownField"]},
        {"IntField": doc2["IntField"], "UnknownField": doc2["UnknownField"]},
        {"IntField": doc3["IntField"]},
    ]
    assert len(results) == len(expected_results)
    assert all(result in expected_results for result in results)


async def test_pagination_asc(prefilled_db: DummyOSDB):
    doc1, doc2, doc3 = prefilled_db.test_docs
    sort = [{"parameter": "IntField", "direction": "asc"}]

    results = await prefilled_db.search(None, [], sort)
    assert results == [doc3, doc2, doc1]

    # Pagination has no effect if a specific page isn't requested
    results = await prefilled_db.search(None, [], sort, per_page=2)
    assert results == [doc3, doc2, doc1]

    results = await prefilled_db.search(None, [], sort, per_page=2, page=1)
    assert results == [doc3, doc2]

    results = await prefilled_db.search(None, [], sort, per_page=2, page=2)
    assert results == [doc1]

    results = await prefilled_db.search(None, [], sort, per_page=2, page=3)
    assert results == []

    results = await prefilled_db.search(None, [], sort, per_page=1, page=1)
    assert results == [doc3]

    results = await prefilled_db.search(None, [], sort, per_page=1, page=2)
    assert results == [doc2]

    results = await prefilled_db.search(None, [], sort, per_page=1, page=3)
    assert results == [doc1]

    results = await prefilled_db.search(None, [], sort, per_page=1, page=4)
    assert results == []


async def test_pagination_desc(prefilled_db: DummyOSDB):
    doc1, doc2, doc3 = prefilled_db.test_docs
    sort = [{"parameter": "IntField", "direction": "desc"}]

    results = await prefilled_db.search(None, [], sort, per_page=2, page=1)
    assert results == [doc1, doc2]

    results = await prefilled_db.search(None, [], sort, per_page=2, page=2)
    assert results == [doc3]


async def test_eq_filter_long(prefilled_db: DummyOSDB):
    doc1, doc2, doc3 = prefilled_db.test_docs
    part = {"parameter": "IntField", "operator": "eq"}

    # Search for an ID which doesn't exist
    results = await prefilled_db.search(None, [part | {"value": "78"}], [])
    assert results == []

    # Check the DB contains what we expect when not filtering
    results = await prefilled_db.search(None, [], [])
    assert len(results) == 3
    assert doc1 in results
    assert doc2 in results
    assert doc3 in results

    # Search separately for the two documents which do exist
    results = await prefilled_db.search(None, [part | {"value": "1234"}], [])
    assert results == [doc1]
    results = await prefilled_db.search(None, [part | {"value": "679"}], [])
    assert results == [doc2]
    results = await prefilled_db.search(None, [part | {"value": "42"}], [])
    assert results == [doc3]


async def test_operators_long(prefilled_db: DummyOSDB):
    doc1, doc2, doc3 = prefilled_db.test_docs
    part = {"parameter": "IntField"}

    query = part | {"operator": "neq", "value": "1234"}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc2["IntField"], doc3["IntField"]}

    query = part | {"operator": "in", "values": ["1234", "42"]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc1["IntField"], doc3["IntField"]}

    query = part | {"operator": "not in", "values": ["1234", "42"]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc2["IntField"]}

    query = part | {"operator": "lt", "value": "1234"}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc2["IntField"], doc3["IntField"]}

    query = part | {"operator": "lt", "value": "679"}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc3["IntField"]}

    query = part | {"operator": "gt", "value": "1234"}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == set()

    query = part | {"operator": "lt", "value": "42"}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == set()


async def test_operators_date(prefilled_db: DummyOSDB):
    doc1, doc2, doc3 = prefilled_db.test_docs
    part = {"parameter": "DateField"}

    query = part | {"operator": "eq", "value": doc3["DateField"]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc3["IntField"]}

    query = part | {"operator": "neq", "value": doc2["DateField"]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc1["IntField"], doc3["IntField"]}

    doc1_time = doc1["DateField"].strftime("%Y-%m-%dT%H:%M")
    doc2_time = doc2["DateField"].strftime("%Y-%m-%dT%H:%M")
    doc3_time = doc3["DateField"].strftime("%Y-%m-%dT%H:%M")

    query = part | {"operator": "in", "values": [doc1_time, doc2_time]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc1["IntField"], doc2["IntField"]}

    query = part | {"operator": "not in", "values": [doc1_time, doc2_time]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc3["IntField"]}

    query = part | {"operator": "lt", "value": doc1_time}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc2["IntField"], doc3["IntField"]}

    query = part | {"operator": "lt", "value": doc3_time}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc2["IntField"]}

    query = part | {"operator": "lt", "value": doc2_time}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == set()

    query = part | {"operator": "gt", "value": doc1_time}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == set()

    query = part | {"operator": "gt", "value": doc3_time}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc1["IntField"]}

    query = part | {"operator": "gt", "value": doc2_time}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc1["IntField"], doc3["IntField"]}


@pytest.mark.parametrize(
    "date_format",
    [
        "%Y-%m-%d",
        "%Y-%m-%dT%H",
        "%Y-%m-%dT%H:%M",
        "%Y-%m-%dT%H:%M:%S",
        "%Y-%m-%dT%H:%M:%S.%fZ",
    ],
)
async def test_operators_date_partial_doc1(prefilled_db: DummyOSDB, date_format: str):
    """Search by datetime without specifying an exact match.

    The parameterized date_format argument should match doc1 but not doc2 or doc3.
    """
    doc1, doc2, doc3 = prefilled_db.test_docs
    formatted_date = doc1["DateField"].strftime(date_format)

    query = {"parameter": "DateField", "operator": "eq", "value": formatted_date}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc1["IntField"]}

    query = {"parameter": "DateField", "operator": "neq", "value": formatted_date}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc2["IntField"], doc3["IntField"]}


async def test_operators_keyword(prefilled_db: DummyOSDB):
    doc1, doc2, doc3 = prefilled_db.test_docs
    part = {"parameter": "KeywordField1"}

    query = part | {"operator": "eq", "value": doc1["KeywordField1"]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc1["IntField"], doc2["IntField"]}

    query = part | {"operator": "neq", "value": doc1["KeywordField1"]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc3["IntField"]}

    part = {"parameter": "KeywordField0"}

    query = part | {
        "operator": "in",
        "values": [doc1["KeywordField0"], doc3["KeywordField0"]],
    }
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc1["IntField"], doc3["IntField"]}

    query = part | {"operator": "in", "values": ["missing"]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == set()

    query = part | {
        "operator": "not in",
        "values": [doc1["KeywordField0"], doc3["KeywordField0"]],
    }
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {doc2["IntField"]}

    query = part | {"operator": "not in", "values": ["missing"]}
    results = await prefilled_db.search(["IntField"], [query], [])
    assert {x["IntField"] for x in results} == {
        doc1["IntField"],
        doc2["IntField"],
        doc3["IntField"],
    }

    # The MockOSDBMixin doesn't validate if types are indexed correctly
    if not isinstance(prefilled_db, MockOSDBMixin):
        with pytest.raises(InvalidQueryError):
            query = part | {"operator": "lt", "value": "a"}
            await prefilled_db.search(["IntField"], [query], [])

        with pytest.raises(InvalidQueryError):
            query = part | {"operator": "gt", "value": "a"}
            await prefilled_db.search(["IntField"], [query], [])


async def test_unknown_operator(prefilled_db: DummyOSDB):
    with pytest.raises(InvalidQueryError):
        await prefilled_db.search(
            None, [{"parameter": "IntField", "operator": "unknown"}], []
        )


async def test_unindexed_field(prefilled_db: DummyOSDB):
    with pytest.raises(InvalidQueryError):
        await prefilled_db.search(
            None,
            [{"parameter": "UnknownField", "operator": "eq", "value": "foobar"}],
            [],
        )


async def test_sort_long(prefilled_db: DummyOSDB):
    doc1, doc2, doc3 = prefilled_db.test_docs
    results = await prefilled_db.search(
        None, [], [{"parameter": "IntField", "direction": "asc"}]
    )
    assert results == [doc3, doc2, doc1]
    results = await prefilled_db.search(
        None, [], [{"parameter": "IntField", "direction": "desc"}]
    )
    assert results == [doc1, doc2, doc3]


async def test_sort_date(prefilled_db: DummyOSDB):
    doc1, doc2, doc3 = prefilled_db.test_docs
    results = await prefilled_db.search(
        None, [], [{"parameter": "DateField", "direction": "asc"}]
    )
    assert results == [doc2, doc3, doc1]
    results = await prefilled_db.search(
        None, [], [{"parameter": "DateField", "direction": "desc"}]
    )
    assert results == [doc1, doc3, doc2]


async def test_sort_keyword(prefilled_db: DummyOSDB):
    doc1, doc2, doc3 = prefilled_db.test_docs
    results = await prefilled_db.search(
        None, [], [{"parameter": "KeywordField0", "direction": "asc"}]
    )
    assert results == [doc1, doc3, doc2]
    results = await prefilled_db.search(
        None, [], [{"parameter": "KeywordField0", "direction": "desc"}]
    )
    assert results == [doc2, doc3, doc1]


async def test_sort_text(prefilled_db: DummyOSDB):
    # The MockOSDBMixin doesn't validate if types are indexed correctly
    if not isinstance(prefilled_db, MockOSDBMixin):
        with pytest.raises(InvalidQueryError):
            await prefilled_db.search(
                None, [], [{"parameter": "TextField", "direction": "asc"}]
            )


async def test_sort_unknown(prefilled_db: DummyOSDB):
    with pytest.raises(InvalidQueryError):
        await prefilled_db.search(
            None, [], [{"parameter": "UnknownField", "direction": "asc"}]
        )


async def test_sort_multiple(prefilled_db: DummyOSDB):
    doc1, doc2, doc3 = prefilled_db.test_docs
    results = await prefilled_db.search(
        None,
        [],
        [
            {"parameter": "KeywordField1", "direction": "asc"},
            {"parameter": "IntField", "direction": "asc"},
        ],
    )
    assert results == [doc2, doc1, doc3]

    results = await prefilled_db.search(
        None,
        [],
        [
            {"parameter": "KeywordField1", "direction": "asc"},
            {"parameter": "IntField", "direction": "desc"},
        ],
    )
    assert results == [doc1, doc2, doc3]

    results = await prefilled_db.search(
        None,
        [],
        [
            {"parameter": "KeywordField1", "direction": "desc"},
            {"parameter": "IntField", "direction": "asc"},
        ],
    )
    assert results == [doc3, doc2, doc1]

    results = await prefilled_db.search(
        None,
        [],
        [
            {"parameter": "IntField", "direction": "asc"},
            {"parameter": "KeywordField1", "direction": "asc"},
        ],
    )
    assert results == [doc3, doc2, doc1]



from __future__ import annotations

from sqlalchemy import (
    BigInteger,
    Index,
    PrimaryKeyConstraint,
    String,
    UniqueConstraint,
)
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from diracx.db.sql.utils import datetime_now, str32, str64, str128, str512


class Base(DeclarativeBase):
    type_annotation_map = {
        str32: String(32),
        str64: String(64),
        str128: String(128),
        str512: String(512),
    }


class SBOwners(Base):
    __tablename__ = "sb_Owners"
    OwnerID: Mapped[int] = mapped_column(autoincrement=True)
    Owner: Mapped[str32]
    OwnerGroup: Mapped[str32]
    VO: Mapped[str64]
    __table_args__ = (
        PrimaryKeyConstraint("OwnerID"),
        UniqueConstraint("Owner", "OwnerGroup", "VO", name="unique_owner_group_vo"),
    )


class SandBoxes(Base):
    __tablename__ = "sb_SandBoxes"
    SBId: Mapped[int] = mapped_column(autoincrement=True)
    OwnerId: Mapped[int]
    SEName: Mapped[str64]
    SEPFN: Mapped[str512]
    Bytes: Mapped[int] = mapped_column(BigInteger)
    RegistrationTime: Mapped[datetime_now]
    LastAccessTime: Mapped[datetime_now]
    Assigned: Mapped[bool] = mapped_column(default=False)
    __table_args__ = (
        PrimaryKeyConstraint("SBId"),
        Index("OwnerId", "OwnerId"),
        UniqueConstraint("SEName", "SEPFN", name="Location"),
    )


class SBEntityMapping(Base):
    __tablename__ = "sb_EntityMapping"
    SBId: Mapped[int]
    EntityId: Mapped[str128]
    Type: Mapped[str64]
    __table_args__ = (
        PrimaryKeyConstraint("SBId", "EntityId", "Type"),
        Index("SBId", "EntityId"),
        UniqueConstraint("SBId", "EntityId", "Type", name="Mapping"),
    )



from __future__ import annotations

from enum import Enum, auto
from typing import Any, Optional
from uuid import UUID

from sqlalchemy import (
    JSON,
    Index,
    String,
    Uuid,
)
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from diracx.db.sql.utils import (
    datetime_now,
    enum_column,
    str128,
    str255,
    str1024,
)

USER_CODE_LENGTH = 8


class Base(DeclarativeBase):
    type_annotation_map = {
        str128: String(128),
        str255: String(255),
        str1024: String(1024),
        dict[str, Any]: JSON,
    }


class FlowStatus(Enum):
    """PENDING -> READY -> DONE.

    Pending is upon insertion
    Ready/Error is set in response to IdP
    Done means the user has been issued the dirac token.
    """

    # The flow is ongoing
    PENDING = auto()
    # The user has been authenticated
    # and we are ready to issue a token
    READY = auto()
    # The token has been minted,
    # the user can no longer login with this flow
    DONE = auto()

    # Something went wrong.
    # Most likely the user did not login in the
    # external IdP
    ERROR = auto()


class DeviceFlows(Base):
    __tablename__ = "DeviceFlows"
    user_code: Mapped[str] = mapped_column(
        "UserCode", String(USER_CODE_LENGTH), primary_key=True
    )
    status: Mapped[FlowStatus] = enum_column(
        "Status", FlowStatus, server_default=FlowStatus.PENDING.name
    )
    creation_time: Mapped[datetime_now] = mapped_column("CreationTime")
    client_id: Mapped[str255] = mapped_column("ClientID")
    scope: Mapped[str1024] = mapped_column("Scope")
    device_code: Mapped[str128] = mapped_column(
        "DeviceCode", unique=True
    )  # Should be a hash
    id_token: Mapped[Optional[dict[str, Any]]] = mapped_column("IDToken")


class AuthorizationFlows(Base):
    __tablename__ = "AuthorizationFlows"
    uuid: Mapped[UUID] = mapped_column("UUID", Uuid(as_uuid=False), primary_key=True)
    status: Mapped[FlowStatus] = enum_column(
        "Status", FlowStatus, server_default=FlowStatus.PENDING.name
    )
    client_id: Mapped[str255] = mapped_column("ClientID")
    creation_time: Mapped[datetime_now] = mapped_column("CreationTime")
    scope: Mapped[str1024] = mapped_column("Scope")
    code_challenge: Mapped[str255] = mapped_column("CodeChallenge")
    code_challenge_method: Mapped[str] = mapped_column("CodeChallengeMethod", String(8))
    redirect_uri: Mapped[str255] = mapped_column("RedirectURI")
    code: Mapped[Optional[str255]] = mapped_column("Code")  # Should be a hash
    id_token: Mapped[Optional[dict[str, Any]]] = mapped_column("IDToken")


class RefreshTokenStatus(Enum):
    """CREATED -> REVOKED.

    Note1: There is no EXPIRED status as it can be calculated from a creation time
    Note2: As part of the refresh token rotation mechanism, the revoked token should be retained
    """

    # The token has been created
    # It does not indicate whether the token is valid or not
    CREATED = auto()

    # The token has been revoked
    REVOKED = auto()


class RefreshTokens(Base):
    """Store attributes bound to a refresh token.

    Also specific user attributes that might be then used to generate access tokens.
    """

    __tablename__ = "RefreshTokens"
    # Refresh token attributes
    jti: Mapped[UUID] = mapped_column("JTI", Uuid(as_uuid=False), primary_key=True)
    status: Mapped[RefreshTokenStatus] = enum_column(
        "Status", RefreshTokenStatus, server_default=RefreshTokenStatus.CREATED.name
    )
    scope: Mapped[str1024] = mapped_column("Scope")

    # User attributes bound to the refresh token
    sub: Mapped[str] = mapped_column("Sub", String(256), index=True)

    __table_args__ = (Index("index_status_sub", status, sub),)



from __future__ import annotations

import pytest
from sqlalchemy.exc import NoResultFound

from diracx.core.exceptions import AuthorizationError
from diracx.db.sql.auth.db import AuthDB
from diracx.db.sql.auth.schema import FlowStatus

MAX_VALIDITY = 2
EXPIRED = 0


@pytest.fixture
async def auth_db(tmp_path):
    auth_db = AuthDB("sqlite+aiosqlite:///:memory:")
    async with auth_db.engine_context():
        async with auth_db.engine.begin() as conn:
            await conn.run_sync(auth_db.metadata.create_all)
        yield auth_db


async def test_insert_id_token(auth_db: AuthDB):
    # First insert
    async with auth_db as auth_db:
        uuid = await auth_db.insert_authorization_flow(
            "client_id", "scope", "code_challenge", "S256", "redirect_uri"
        )

    id_token = {"sub": "myIdToken"}

    async with auth_db as auth_db:
        with pytest.raises(AuthorizationError):
            code, redirect_uri = await auth_db.authorization_flow_insert_id_token(
                uuid, id_token, EXPIRED
            )
        code, redirect_uri = await auth_db.authorization_flow_insert_id_token(
            uuid, id_token, MAX_VALIDITY
        )
        assert redirect_uri == "redirect_uri"

    # Cannot add a id_token a second time
    async with auth_db as auth_db:
        with pytest.raises(AuthorizationError):
            await auth_db.authorization_flow_insert_id_token(
                uuid, id_token, MAX_VALIDITY
            )

    async with auth_db as auth_db:
        with pytest.raises(NoResultFound):
            await auth_db.get_authorization_flow(code, EXPIRED)
        res = await auth_db.get_authorization_flow(code, MAX_VALIDITY)
        assert res["IDToken"] == id_token

    # Cannot add a id_token after finishing the flow
    async with auth_db as auth_db:
        with pytest.raises(AuthorizationError):
            await auth_db.authorization_flow_insert_id_token(
                uuid, id_token, MAX_VALIDITY
            )


async def test_insert(auth_db: AuthDB):
    # First insert
    async with auth_db as auth_db:
        uuid1 = await auth_db.insert_authorization_flow(
            "client_id", "scope", "code_challenge", "S256", "redirect_uri"
        )
        uuid2 = await auth_db.insert_authorization_flow(
            "client_id2",
            "scope2",
            "code_challenge2",
            "S256",
            "redirect_uri2",
        )

    assert uuid1 != uuid2


async def test_clean_authorization_flows(auth_db: AuthDB):
    # Insert two authorization flows
    async with auth_db as auth_db:
        _ = await auth_db.insert_authorization_flow(
            "client_id", "scope", "code_challenge", "S256", "redirect_uri"
        )
        uuid2 = await auth_db.insert_authorization_flow(
            "client_id2", "scope2", "code_challenge2", "S256", "redirect_uri2"
        )
        uuid3 = await auth_db.insert_authorization_flow(
            "client_id3", "scope3", "code_challenge3", "S256", "redirect_uri3"
        )
        uuid4 = await auth_db.insert_authorization_flow(
            "client_id4", "scope4", "code_challenge4", "S256", "redirect_uri4"
        )

    id_token = {"sub": "myIdToken"}

    async with auth_db as auth_db:
        _, _ = await auth_db.authorization_flow_insert_id_token(uuid2, id_token, 1)
        code3, _ = await auth_db.authorization_flow_insert_id_token(uuid3, id_token, 1)
        code4, _ = await auth_db.authorization_flow_insert_id_token(uuid4, id_token, 1)

    async with auth_db as auth_db:
        await auth_db.update_authorization_flow_status(code3, FlowStatus.DONE)
        await auth_db.update_authorization_flow_status(code4, FlowStatus.ERROR)

    # Check the number of deleted authorization flow (should be 0)
    async with auth_db as auth_db:
        deleted_auth = await auth_db.clean_expired_authorization_flows(max_retention=30)
        assert deleted_auth == 0

    # Check the number of deleted authorization flow (should be 4: 1 PENDING, 1 READY, 1 DONE, 1 ERROR)
    async with auth_db as auth_db:
        deleted_auth = await auth_db.clean_expired_authorization_flows(max_retention=0)
        assert deleted_auth == 4

    # Check the number of deleted authorization flow (should be 0 because there is nothing left to delete)
    async with auth_db as auth_db:
        deleted_auth = await auth_db.clean_expired_authorization_flows(max_retention=0)
        assert deleted_auth == 0



from __future__ import annotations

from datetime import UTC, datetime

from sqlalchemy import Numeric, PrimaryKeyConstraint, String, TypeDecorator
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from ..utils import datetime_now, str32, str128, str255


class JobLoggingDBBase(DeclarativeBase):
    type_annotation_map = {
        str32: String(32),
        str128: String(128),
        str255: String(255),
    }


class MagicEpochDateTime(TypeDecorator):
    """A SQLAlchemy type to store a datetime as a numeric value.

    Representing the seconds elapsed since MAGIC_EPOC_NUMBER. The underlying column is defined as
    Numeric(12,3) which provides a fixed-precision representation.
    """

    impl = Numeric(12, 3)
    cache_ok = True

    MAGIC_EPOC_NUMBER = 1270000000

    def process_bind_param(self, value, dialect):
        """Convert a Python datetime to a numeric value: (timestamp - MAGIC_EPOC_NUMBER).

        The result is rounded to three decimal places.
        """
        if value is None:
            return None
        if isinstance(value, datetime):
            # Convert datetime to seconds since the Unix epoch, subtract our magic epoch,
            # and round to three decimal places.
            epoch_seconds = (
                value.replace(tzinfo=UTC).timestamp() - self.MAGIC_EPOC_NUMBER
            )
            return round(epoch_seconds, 3)
        raise ValueError(
            "Expected a datetime object for MagicEpochDateTime bind parameter."
        )

    def process_result_value(self, value, dialect):
        """Convert the numeric database value back into a Python datetime.

        Reversing the stored difference (adding MAGIC_EPOC_NUMBER).
        """
        if value is None:
            return None
        # Carefully convert from Decimal to datetime to avoid losing precision
        value += self.MAGIC_EPOC_NUMBER
        value_int = int(value)
        result = datetime.fromtimestamp(value_int, tz=UTC)
        return result.replace(microsecond=int((value - value_int) * 1_000_000))


class LoggingInfo(JobLoggingDBBase):
    __tablename__ = "LoggingInfo"
    job_id: Mapped[int] = mapped_column("JobID")
    seq_num: Mapped[int] = mapped_column("SeqNum")
    status: Mapped[str32] = mapped_column("Status", default="")
    minor_status: Mapped[str128] = mapped_column("MinorStatus", default="")
    application_status: Mapped[str255] = mapped_column("ApplicationStatus", default="")
    status_time: Mapped[datetime_now] = mapped_column("StatusTime")
    status_time_order: Mapped[datetime] = mapped_column(
        "StatusTimeOrder", MagicEpochDateTime(), default=0
    )
    source: Mapped[str32] = mapped_column("StatusSource", default="Unknown")
    __table_args__ = (PrimaryKeyConstraint("JobID", "SeqNum"),)



from __future__ import annotations

import secrets
from datetime import timezone

import pytest
from sqlalchemy.exc import NoResultFound

from diracx.core.exceptions import AuthorizationError
from diracx.db.sql.auth.db import AuthDB
from diracx.db.sql.auth.schema import USER_CODE_LENGTH, FlowStatus
from diracx.db.sql.utils.functions import substract_date

MAX_VALIDITY = 2
EXPIRED = 0


@pytest.fixture
async def auth_db(tmp_path):
    auth_db = AuthDB("sqlite+aiosqlite:///:memory:")
    async with auth_db.engine_context():
        async with auth_db.engine.begin() as conn:
            await conn.run_sync(auth_db.metadata.create_all)
        yield auth_db


async def test_device_user_code_collision(auth_db: AuthDB, monkeypatch):
    monkeypatch.setattr(secrets, "choice", lambda _: "A")

    # First insert should work
    async with auth_db as auth_db:
        code, device = await auth_db.insert_device_flow(
            "client_id",
            "scope",
        )
        assert code == "A" * USER_CODE_LENGTH
        assert device

    async with auth_db as auth_db:
        with pytest.raises(NotImplementedError, match="insert new device flow"):
            await auth_db.insert_device_flow("client_id", "scope")

    monkeypatch.setattr(secrets, "choice", lambda _: "B")

    async with auth_db as auth_db:
        code, device = await auth_db.insert_device_flow(
            "client_id",
            "scope",
        )
        assert code == "B" * USER_CODE_LENGTH
        assert device


async def test_device_flow_lookup(auth_db: AuthDB, monkeypatch):
    async with auth_db as auth_db:
        with pytest.raises(NoResultFound):
            await auth_db.device_flow_validate_user_code("NotInserted", MAX_VALIDITY)

    async with auth_db as auth_db:
        with pytest.raises(NoResultFound):
            await auth_db.get_device_flow("NotInserted")

    # First insert
    async with auth_db as auth_db:
        user_code1, device_code1 = await auth_db.insert_device_flow(
            "client_id1",
            "scope1",
        )
        user_code2, device_code2 = await auth_db.insert_device_flow(
            "client_id2",
            "scope2",
        )

        assert user_code1 != user_code2

    async with auth_db as auth_db:
        with pytest.raises(NoResultFound):
            await auth_db.device_flow_validate_user_code(user_code1, EXPIRED)

        await auth_db.device_flow_validate_user_code(user_code1, MAX_VALIDITY)
        await auth_db.device_flow_validate_user_code(user_code2, MAX_VALIDITY)

    async with auth_db as auth_db:
        with pytest.raises(AuthorizationError):
            await auth_db.device_flow_insert_id_token(
                user_code1, {"token": "mytoken"}, EXPIRED
            )

        await auth_db.device_flow_insert_id_token(
            user_code1, {"token": "mytoken"}, MAX_VALIDITY
        )

        # We should not be able to insert a id_token a second time
        with pytest.raises(AuthorizationError):
            await auth_db.device_flow_insert_id_token(
                user_code1, {"token": "mytoken2"}, MAX_VALIDITY
            )

        res = await auth_db.get_device_flow(device_code1)
        # The device code should be expired
        assert res["CreationTime"].replace(tzinfo=timezone.utc) > substract_date(
            seconds=MAX_VALIDITY
        )

        res = await auth_db.get_device_flow(device_code1)
        assert res["UserCode"] == user_code1
        assert res["IDToken"] == {"token": "mytoken"}

    # Re-adding a token should not work after it's been minted
    async with auth_db as auth_db:
        with pytest.raises(AuthorizationError):
            await auth_db.device_flow_insert_id_token(
                user_code1, {"token": "mytoken"}, MAX_VALIDITY
            )


async def test_device_flow_insert_id_token(auth_db: AuthDB):
    # First insert
    async with auth_db as auth_db:
        user_code, device_code = await auth_db.insert_device_flow(
            "client_id",
            "scope",
        )

    # Make sure it exists, and is Pending
    async with auth_db as auth_db:
        await auth_db.device_flow_validate_user_code(user_code, MAX_VALIDITY)

    id_token = {"sub": "myIdToken"}

    async with auth_db as auth_db:
        await auth_db.device_flow_insert_id_token(user_code, id_token, MAX_VALIDITY)

    # The user code has been invalidated
    async with auth_db as auth_db:
        with pytest.raises(NoResultFound):
            await auth_db.device_flow_validate_user_code(user_code, MAX_VALIDITY)

    async with auth_db as auth_db:
        res = await auth_db.get_device_flow(device_code)
        assert res["IDToken"] == id_token


async def test_clean_device_flows(auth_db: AuthDB):
    # Insert two device flows
    async with auth_db as auth_db:
        _, _ = await auth_db.insert_device_flow("client_id1", "scope1")
        user_code2, _ = await auth_db.insert_device_flow("client_id2", "scope2")
        _, device_code3 = await auth_db.insert_device_flow("client_id3", "scope3")
        _, device_code4 = await auth_db.insert_device_flow("client_id4", "scope4")

    async with auth_db as auth_db:
        await auth_db.device_flow_validate_user_code(user_code2, MAX_VALIDITY)
        await auth_db.device_flow_insert_id_token(
            user_code2, {"sub": "myIdToken"}, MAX_VALIDITY
        )
        await auth_db.update_device_flow_status(device_code3, FlowStatus.DONE)
        await auth_db.update_device_flow_status(device_code4, FlowStatus.ERROR)

    # Check the number of deleted device flows (should be 0)
    async with auth_db as auth_db:
        deleted_device = await auth_db.clean_expired_device_flows(max_retention=30)
        assert deleted_device == 0

    # Check the number of deleted device flows (should be 4: 1 PENDING, 1 READY, 1 DONE, 1 ERROR)
    async with auth_db as auth_db:
        deleted_device = await auth_db.clean_expired_device_flows(max_retention=0)
        assert deleted_device == 4

    # Check the number of deleted device flow (should be 0 because there is nothing left to delete)
    async with auth_db as auth_db:
        deleted_device = await auth_db.clean_expired_device_flows(max_retention=0)
        assert deleted_device == 0



from __future__ import annotations

import logging
import secrets
from datetime import UTC, datetime
from itertools import pairwise

from dateutil.rrule import MONTHLY, rrule
from sqlalchemy import delete, insert, select, text, update
from sqlalchemy.exc import IntegrityError, NoResultFound
from sqlalchemy.ext.asyncio import AsyncConnection
from uuid_utils import UUID, uuid7

from diracx.core.exceptions import (
    AuthorizationError,
    TokenNotFoundError,
)
from diracx.db.sql.utils import BaseSQLDB, hash, substract_date, uuid7_from_datetime

from .schema import (
    AuthorizationFlows,
    DeviceFlows,
    FlowStatus,
    RefreshTokens,
    RefreshTokenStatus,
)
from .schema import Base as AuthDBBase

# https://datatracker.ietf.org/doc/html/rfc8628#section-6.1
USER_CODE_ALPHABET = "BCDFGHJKLMNPQRSTVWXZ"
MAX_RETRY = 5

logger = logging.getLogger(__name__)


class AuthDB(BaseSQLDB):
    metadata = AuthDBBase.metadata

    @classmethod
    async def post_create(cls, conn: AsyncConnection) -> None:
        """Create partitions.

        If it is a MySQL DB and it does not have
        it yet and the table does not have any data yet.
        We do this as a post_create step as sqlalchemy does not support
        partition so well.
        """
        if conn.dialect.name == "mysql":
            check_partition_query = text(
                "SELECT PARTITION_NAME FROM information_schema.partitions "
                "WHERE TABLE_NAME = 'RefreshTokens' AND PARTITION_NAME is not NULL"
            )
            partition_names = (await conn.execute(check_partition_query)).all()

            if not partition_names:
                # Create a monthly partition from today until 2 years
                # The partition are named p_<year>_<month>
                start_date = datetime.now(tz=UTC).replace(
                    day=1, hour=0, minute=0, second=0, microsecond=0
                )
                end_date = start_date.replace(year=start_date.year + 2)

                dates = [
                    dt for dt in rrule(MONTHLY, dtstart=start_date, until=end_date)
                ]

                partition_list = []
                for name, limit in pairwise(dates):
                    partition_list.append(
                        f"PARTITION p_{name.year}_{name.month} "
                        f"VALUES LESS THAN ('{str(uuid7_from_datetime(limit, randomize=False)).replace('-', '')}')"
                    )
                partition_list.append("PARTITION p_future VALUES LESS THAN (MAXVALUE)")

                alter_query = text(
                    f"ALTER TABLE RefreshTokens PARTITION BY RANGE COLUMNS (JTI) ({','.join(partition_list)})"
                )

                check_table_empty_query = text("SELECT * FROM RefreshTokens LIMIT 1")
                refresh_table_content = (
                    await conn.execute(check_table_empty_query)
                ).all()
                if refresh_table_content:
                    logger.warning(
                        "RefreshTokens table not empty. Run the following query yourself"
                    )
                    logger.warning(alter_query)
                    return

                await conn.execute(alter_query)

                partition_names = (
                    await conn.execute(
                        check_partition_query, {"table_name": "RefreshTokens"}
                    )
                ).all()
                assert partition_names, (
                    f"There should be partitions now {partition_names}"
                )

    async def device_flow_validate_user_code(
        self, user_code: str, max_validity: int
    ) -> str:
        """Validate that the user_code can be used (Pending status, not expired).

        Returns the scope field for the given user_code

        :raises:
            NoResultFound if no such user code currently Pending
        """
        stmt = select(DeviceFlows.scope).where(
            DeviceFlows.user_code == user_code,
            DeviceFlows.status == FlowStatus.PENDING,
            DeviceFlows.creation_time > substract_date(seconds=max_validity),
        )

        return (await self.conn.execute(stmt)).scalar_one()

    async def get_device_flow(self, device_code: str):
        """raise: NoResultFound."""
        # The with_for_update
        # prevents that the token is retrieved
        # multiple time concurrently
        stmt = select(DeviceFlows).with_for_update()
        stmt = stmt.where(
            DeviceFlows.device_code == hash(device_code),
        )
        return dict((await self.conn.execute(stmt)).one()._mapping)

    async def update_device_flow_status(
        self, device_code: str, status: FlowStatus
    ) -> None:
        stmt = update(DeviceFlows).where(
            DeviceFlows.device_code == hash(device_code),
        )
        stmt = stmt.values(status=status)
        await self.conn.execute(stmt)

    async def device_flow_insert_id_token(
        self, user_code: str, id_token: dict[str, str], max_validity: int
    ) -> None:
        """raise: AuthorizationError if no such code or status not pending."""
        stmt = update(DeviceFlows)
        stmt = stmt.where(
            DeviceFlows.user_code == user_code,
            DeviceFlows.status == FlowStatus.PENDING,
            DeviceFlows.creation_time > substract_date(seconds=max_validity),
        )
        stmt = stmt.values(id_token=id_token, status=FlowStatus.READY)
        res = await self.conn.execute(stmt)
        if res.rowcount != 1:
            raise AuthorizationError(
                f"{res.rowcount} rows matched user_code {user_code}"
            )

    async def insert_device_flow(
        self,
        client_id: str,
        scope: str,
    ) -> tuple[str, str]:
        # Because the user_code might be short, there is a risk of conflicts
        # This is why we retry multiple times
        for _ in range(MAX_RETRY):
            user_code = "".join(
                secrets.choice(USER_CODE_ALPHABET)
                for _ in range(DeviceFlows.user_code.type.length)
            )
            device_code = secrets.token_urlsafe()

            # Hash the the device_code to avoid leaking information
            hashed_device_code = hash(device_code)

            stmt = insert(DeviceFlows).values(
                client_id=client_id,
                scope=scope,
                user_code=user_code,
                device_code=hashed_device_code,
            )
            try:
                await self.conn.execute(stmt)

            except IntegrityError:
                logger.warning(
                    "Device flow code collision detected, retrying (user_code=%s)",
                    user_code,
                )
                continue

            return user_code, device_code
        raise NotImplementedError(
            f"Could not insert new device flow after {MAX_RETRY} retries"
        )

    async def insert_authorization_flow(
        self,
        client_id: str,
        scope: str,
        code_challenge: str,
        code_challenge_method: str,
        redirect_uri: str,
    ) -> str:
        uuid = str(uuid7())

        stmt = insert(AuthorizationFlows).values(
            uuid=uuid,
            client_id=client_id,
            scope=scope,
            code_challenge=code_challenge,
            code_challenge_method=code_challenge_method,
            redirect_uri=redirect_uri,
        )

        await self.conn.execute(stmt)

        return uuid

    async def authorization_flow_insert_id_token(
        self, uuid: str, id_token: dict[str, str], max_validity: int
    ) -> tuple[str, str]:
        """Return code, redirect_uri.

        :raises: AuthorizationError if no such uuid or status not pending.
        """
        # Hash the code to avoid leaking information
        code = secrets.token_urlsafe()
        hashed_code = hash(code)

        stmt = update(AuthorizationFlows)

        stmt = stmt.where(
            AuthorizationFlows.uuid == uuid,
            AuthorizationFlows.status == FlowStatus.PENDING,
            AuthorizationFlows.creation_time > substract_date(seconds=max_validity),
        )

        stmt = stmt.values(id_token=id_token, code=hashed_code, status=FlowStatus.READY)
        res = await self.conn.execute(stmt)

        if res.rowcount != 1:
            raise AuthorizationError(f"{res.rowcount} rows matched uuid {uuid}")

        stmt = select(AuthorizationFlows.code, AuthorizationFlows.redirect_uri)
        stmt = stmt.where(AuthorizationFlows.uuid == uuid)
        row = (await self.conn.execute(stmt)).one()
        return code, row.RedirectURI

    async def get_authorization_flow(self, code: str, max_validity: int):
        """Get the authorization flow details based on the code."""
        hashed_code = hash(code)
        # The with_for_update
        # prevents that the token is retrieved
        # multiple time concurrently
        stmt = select(AuthorizationFlows).with_for_update()
        stmt = stmt.where(
            AuthorizationFlows.code == hashed_code,
            AuthorizationFlows.creation_time > substract_date(seconds=max_validity),
        )

        return dict((await self.conn.execute(stmt)).one()._mapping)

    async def update_authorization_flow_status(
        self, code: str, status: FlowStatus
    ) -> None:
        """Update the status of an authorization flow based on the code."""
        hashed_code = hash(code)
        await self.conn.execute(
            update(AuthorizationFlows)
            .where(AuthorizationFlows.code == hashed_code)
            .values(status=status)
        )

    async def insert_refresh_token(
        self,
        jti: UUID,
        subject: str,
        scope: str,
    ) -> None:
        """Insert a refresh token in the DB.

        As well as user attributes required to generate access tokens.
        """
        # Insert values into the DB
        stmt = insert(RefreshTokens).values(
            jti=str(jti),
            sub=subject,
            scope=scope,
        )
        await self.conn.execute(stmt)

    async def get_refresh_token(self, jti: UUID) -> dict:
        """Get refresh token details bound to a given JWT ID."""
        jti = str(jti)
        # The with_for_update
        # prevents that the token is retrieved
        # multiple time concurrently
        stmt = select(RefreshTokens).with_for_update()
        stmt = stmt.where(
            RefreshTokens.jti == jti,
        )
        try:
            res = dict((await self.conn.execute(stmt)).one()._mapping)
        except NoResultFound as e:
            raise TokenNotFoundError(jti) from e

        return res

    async def get_user_refresh_tokens(self, subject: str | None = None) -> list[dict]:
        """Get a list of refresh token details based on a subject ID (not revoked)."""
        # Get a list of refresh tokens
        stmt = select(RefreshTokens).with_for_update()

        if subject:
            stmt = stmt.where(
                RefreshTokens.sub == subject,
                RefreshTokens.status != RefreshTokenStatus.REVOKED,
            )

        res = (await self.conn.execute(stmt)).all()

        # Convert the results into dict
        refresh_tokens = []
        for refresh_token in res:
            refresh_tokens.append(dict(refresh_token._mapping))

        return refresh_tokens

    async def revoke_refresh_token(self, jti: UUID):
        """Revoke a token given by its JWT ID."""
        await self.conn.execute(
            update(RefreshTokens)
            .where(RefreshTokens.jti == str(jti))
            .values(status=RefreshTokenStatus.REVOKED)
        )

    async def revoke_user_refresh_tokens(self, subject):
        """Revoke all the refresh tokens belonging to a user (subject ID)."""
        await self.conn.execute(
            update(RefreshTokens)
            .where(RefreshTokens.sub == subject)
            .values(status=RefreshTokenStatus.REVOKED)
        )

    async def clean_expired_refresh_tokens(self, max_validity: int) -> int:
        """Delete expired refresh tokens.

        max_validity: Maximum validity time in minutes for refresh tokens.
        """
        expired_date = str(
            uuid7_from_datetime(substract_date(minutes=max_validity), randomize=False)
        )
        stmt_expired = delete(RefreshTokens).where(
            RefreshTokens.status == RefreshTokenStatus.CREATED,
            RefreshTokens.jti < expired_date,
        )
        res_expired = await self.conn.execute(stmt_expired)

        return res_expired.rowcount

    async def clean_revoked_refresh_tokens(self, max_retention: int) -> int:
        """Delete old revoked refresh tokens.

        max_retention: Maximum retention time in minutes for revoked refresh tokens.
        """
        revoked_date = str(
            uuid7_from_datetime(substract_date(minutes=max_retention), randomize=False)
        )
        stmt_revoked = delete(RefreshTokens).where(
            RefreshTokens.status == RefreshTokenStatus.REVOKED,
            RefreshTokens.jti < revoked_date,
        )
        res_revoked = await self.conn.execute(stmt_revoked)

        return res_revoked.rowcount

    async def clean_expired_authorization_flows(self, max_retention: int) -> int:
        """Delete old authorization flows.

        max_retention: Maximum retention time in minutes for expired authorization flows.
        Must be bigger than authorization_flow_expiration_seconds.
        """
        stmt_auth = delete(AuthorizationFlows).where(
            AuthorizationFlows.creation_time < substract_date(minutes=max_retention),
        )
        res_auth = await self.conn.execute(stmt_auth)

        return res_auth.rowcount

    async def clean_expired_device_flows(self, max_retention: int) -> int:
        """Delete old device flows.

        max_retention: Maximum retention time in minutes for expired device flows.
        Must be bigger than device_flow_expiration_seconds.
        """
        stmt_device = delete(DeviceFlows).where(
            DeviceFlows.creation_time < substract_date(minutes=max_retention),
        )
        res_device = await self.conn.execute(stmt_device)

        return res_device.rowcount



from __future__ import annotations

__all__ = [
    "_get_columns",
    "utcnow",
    "BaseSQLDB",
    "EnumBackedBool",
    "enum_column",
    "apply_search_filters",
    "apply_sort_constraints",
    "substract_date",
    "hash",
    "SQLDBUnavailableError",
    "uuid7_from_datetime",
    "uuid7_to_datetime",
    "datetime_now",
    "str32",
    "str64",
    "str128",
    "str255",
    "str512",
    "str1024",
]

from .base import (
    BaseSQLDB,
    SQLDBUnavailableError,
    _get_columns,
    apply_search_filters,
    apply_sort_constraints,
    uuid7_from_datetime,
    uuid7_to_datetime,
)
from .functions import hash, substract_date, utcnow
from .types import (
    EnumBackedBool,
    datetime_now,
    enum_column,
    str32,
    str64,
    str128,
    str255,
    str512,
    str1024,
)



from __future__ import annotations

from sqlalchemy import insert
from uuid_utils import UUID

from diracx.db.sql.utils import BaseSQLDB

from .schema import Base as DummyDBBase
from .schema import Cars, Owners


class DummyDB(BaseSQLDB):
    """Illustrate some important aspect of writing DB classes in DiracX.

    It is mostly pure SQLAlchemy, with a few convention

    Document the secrets
    """

    # This needs to be here for the BaseSQLDB to create the engine
    metadata = DummyDBBase.metadata

    async def summary(self, group_by, search) -> list[dict[str, str | int]]:
        return await self._summary(Cars, group_by, search)

    async def insert_owner(self, name: str) -> int:
        stmt = insert(Owners).values(name=name)
        result = await self.conn.execute(stmt)
        # await self.engine.commit()
        return result.lastrowid

    async def insert_car(self, license_plate: UUID, model: str, owner_id: int) -> int:
        stmt = insert(Cars).values(
            license_plate=license_plate, model=model, owner_id=owner_id
        )

        result = await self.conn.execute(stmt)
        # await self.engine.commit()
        return result.lastrowid



from __future__ import annotations

from datetime import datetime
from typing import Optional

import sqlalchemy.types as types
from sqlalchemy import (
    ForeignKey,
    Index,
    String,
    Text,
)
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column
from typing_extensions import Annotated

from diracx.db.sql.utils.types import SmarterDateTime

from ..utils import EnumBackedBool, str32, str64, str128, str255

str100 = Annotated[str, 100]


class JobDBBase(DeclarativeBase):
    type_annotation_map = {
        str32: String(32),
        str64: String(64),
        str100: String(100),
        str128: String(128),
        str255: String(255),
    }


class AccountedFlagEnum(types.TypeDecorator):
    """Maps a ``AccountedFlagEnum()`` column to True/False in Python."""

    impl = types.Enum("True", "False", "Failed", name="accounted_flag_enum")
    cache_ok = True

    def process_bind_param(self, value, dialect) -> str:
        if value is True:
            return "True"
        elif value is False:
            return "False"
        elif value == "Failed":
            return "Failed"
        else:
            raise NotImplementedError(value, dialect)

    def process_result_value(self, value, dialect) -> bool | str:
        if value == "True":
            return True
        elif value == "False":
            return False
        elif value == "Failed":
            return "Failed"
        else:
            raise NotImplementedError(f"Unknown {value=}")


class Jobs(JobDBBase):
    __tablename__ = "Jobs"

    job_id: Mapped[int] = mapped_column(
        "JobID",
        ForeignKey("JobJDLs.JobID", ondelete="CASCADE"),
        primary_key=True,
        default=0,
    )
    job_type: Mapped[str32] = mapped_column("JobType", default="user")
    job_group: Mapped[str32] = mapped_column("JobGroup", default="00000000")
    site: Mapped[str100] = mapped_column("Site", default="ANY")
    job_name: Mapped[str128] = mapped_column("JobName", default="Unknown")
    owner: Mapped[str64] = mapped_column("Owner", default="Unknown")
    owner_group: Mapped[str128] = mapped_column("OwnerGroup", default="Unknown")
    vo: Mapped[str32] = mapped_column("VO")
    submission_time: Mapped[Optional[datetime]] = mapped_column(
        "SubmissionTime",
        SmarterDateTime(),
    )
    reschedule_time: Mapped[Optional[datetime]] = mapped_column(
        "RescheduleTime",
        SmarterDateTime(),
    )
    last_update_time: Mapped[Optional[datetime]] = mapped_column(
        "LastUpdateTime",
        SmarterDateTime(),
    )
    start_exec_time: Mapped[Optional[datetime]] = mapped_column(
        "StartExecTime",
        SmarterDateTime(),
    )
    heart_beat_time: Mapped[Optional[datetime]] = mapped_column(
        "HeartBeatTime",
        SmarterDateTime(),
    )
    end_exec_time: Mapped[Optional[datetime]] = mapped_column(
        "EndExecTime",
        SmarterDateTime(),
    )
    status: Mapped[str32] = mapped_column("Status", default="Received")
    minor_status: Mapped[str128] = mapped_column("MinorStatus", default="Unknown")
    application_status: Mapped[str255] = mapped_column(
        "ApplicationStatus", default="Unknown"
    )
    user_priority: Mapped[int] = mapped_column("UserPriority", default=0)
    reschedule_counter: Mapped[int] = mapped_column("RescheduleCounter", default=0)
    verified_flag: Mapped[bool] = mapped_column(
        "VerifiedFlag", EnumBackedBool(), default=False
    )
    accounted_flag: Mapped[bool | str] = mapped_column(
        "AccountedFlag", AccountedFlagEnum(), default=False
    )

    __table_args__ = (
        Index("JobType", "JobType"),
        Index("JobGroup", "JobGroup"),
        Index("Site", "Site"),
        Index("Owner", "Owner"),
        Index("OwnerGroup", "OwnerGroup"),
        Index("Status", "Status"),
        Index("MinorStatus", "MinorStatus"),
        Index("ApplicationStatus", "ApplicationStatus"),
        Index("StatusSite", "Status", "Site"),
        Index("LastUpdateTime", "LastUpdateTime"),
    )


class JobJDLs(JobDBBase):
    __tablename__ = "JobJDLs"
    job_id: Mapped[int] = mapped_column("JobID", autoincrement=True, primary_key=True)
    jdl: Mapped[str] = mapped_column("JDL", Text)
    job_requirements: Mapped[str] = mapped_column("JobRequirements", Text)
    original_jdl: Mapped[str] = mapped_column("OriginalJDL", Text)


class InputData(JobDBBase):
    __tablename__ = "InputData"
    job_id: Mapped[int] = mapped_column(
        "JobID", ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    lfn: Mapped[str255] = mapped_column("LFN", default="", primary_key=True)
    status: Mapped[str32] = mapped_column("Status", default="AprioriGood")


class JobParameters(JobDBBase):
    __tablename__ = "JobParameters"
    job_id: Mapped[int] = mapped_column(
        "JobID", ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    name: Mapped[str100] = mapped_column("Name", primary_key=True)
    value: Mapped[str] = mapped_column("Value", Text)


class OptimizerParameters(JobDBBase):
    __tablename__ = "OptimizerParameters"
    job_id: Mapped[int] = mapped_column(
        "JobID", ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    name: Mapped[str100] = mapped_column("Name", primary_key=True)
    value: Mapped[str] = mapped_column("Value", Text)


class AtticJobParameters(JobDBBase):
    __tablename__ = "AtticJobParameters"
    job_id: Mapped[int] = mapped_column(
        "JobID", ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    name: Mapped[str100] = mapped_column("Name", primary_key=True)
    value: Mapped[str] = mapped_column("Value", Text)
    reschedule_cycle: Mapped[int] = mapped_column("RescheduleCycle")


class HeartBeatLoggingInfo(JobDBBase):
    __tablename__ = "HeartBeatLoggingInfo"
    job_id: Mapped[int] = mapped_column(
        "JobID", ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    name: Mapped[str100] = mapped_column("Name", primary_key=True)
    value: Mapped[str] = mapped_column("Value", Text)
    heart_beat_time: Mapped[datetime] = mapped_column(
        "HeartBeatTime",
        SmarterDateTime(),
        primary_key=True,
    )


class JobCommands(JobDBBase):
    __tablename__ = "JobCommands"
    job_id: Mapped[int] = mapped_column(
        "JobID", ForeignKey("Jobs.JobID", ondelete="CASCADE"), primary_key=True
    )
    command: Mapped[str100] = mapped_column("Command")
    arguments: Mapped[str100] = mapped_column("Arguments")
    status: Mapped[str64] = mapped_column("Status", default="Received")
    reception_time: Mapped[datetime] = mapped_column(
        "ReceptionTime",
        SmarterDateTime(),
        primary_key=True,
    )
    execution_time: Mapped[Optional[datetime]] = mapped_column(
        "ExecutionTime",
        SmarterDateTime(),
    )



# The utils class define some boilerplate types that should be used
# in place of the SQLAlchemy one. Have a look at them
from __future__ import annotations

from uuid import UUID

from sqlalchemy import ForeignKey, String
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from diracx.db.sql.utils import datetime_now, str255


class Base(DeclarativeBase):
    type_annotation_map = {
        str255: String(255),
    }


class Owners(Base):
    __tablename__ = "Owners"
    owner_id: Mapped[int] = mapped_column(
        "OwnerID", primary_key=True, autoincrement=True
    )
    creation_time: Mapped[datetime_now] = mapped_column("CreationTime")
    name: Mapped[str255] = mapped_column("Name")


class Cars(Base):
    __tablename__ = "Cars"
    license_plate: Mapped[UUID] = mapped_column("LicensePlate", primary_key=True)
    model: Mapped[str255] = mapped_column("Model")
    owner_id: Mapped[int] = mapped_column("OwnerID", ForeignKey(Owners.owner_id))



from __future__ import annotations

from sqlalchemy import (
    BigInteger,
    ForeignKey,
    Index,
    String,
)
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column

from diracx.db.sql.utils import (
    str32,
    str64,
    str128,
    str255,
)


class TaskQueueDBBase(DeclarativeBase):
    type_annotation_map = {
        str32: String(32),
        str64: String(64),
        str128: String(128),
        str255: String(255),
    }


class TaskQueues(TaskQueueDBBase):
    __tablename__ = "tq_TaskQueues"
    TQId: Mapped[int] = mapped_column(primary_key=True)
    Owner: Mapped[str255]
    OwnerGroup: Mapped[str32]
    VO: Mapped[str32]
    CPUTime: Mapped[int] = mapped_column(BigInteger)
    Priority: Mapped[float]
    Enabled: Mapped[bool] = mapped_column(default=0)
    __table_args__ = (Index("TQOwner", "Owner", "OwnerGroup", "CPUTime"),)


class JobsQueue(TaskQueueDBBase):
    __tablename__ = "tq_Jobs"
    TQId: Mapped[int] = mapped_column(
        ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    JobId: Mapped[int] = mapped_column(primary_key=True)
    Priority: Mapped[int]
    RealPriority: Mapped[float]
    __table_args__ = (Index("TaskIndex", "TQId"),)


class SitesQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToSites"
    TQId: Mapped[int] = mapped_column(
        ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value: Mapped[str64] = mapped_column(primary_key=True)
    __table_args__ = (
        Index("SitesTaskIndex", "TQId"),
        Index("SitesIndex", "Value"),
    )


class GridCEsQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToGridCEs"
    TQId: Mapped[int] = mapped_column(
        ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value: Mapped[str64] = mapped_column(primary_key=True)
    __table_args__ = (
        Index("GridCEsTaskIndex", "TQId"),
        Index("GridCEsValueIndex", "Value"),
    )


class BannedSitesQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToBannedSites"
    TQId: Mapped[int] = mapped_column(
        ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value: Mapped[str64] = mapped_column(primary_key=True)
    __table_args__ = (
        Index("BannedSitesTaskIndex", "TQId"),
        Index("BannedSitesValueIndex", "Value"),
    )


class PlatformsQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToPlatforms"
    TQId: Mapped[int] = mapped_column(
        ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value: Mapped[str64] = mapped_column(primary_key=True)
    __table_args__ = (
        Index("PlatformsTaskIndex", "TQId"),
        Index("PlatformsValueIndex", "Value"),
    )


class JobTypesQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToJobTypes"
    TQId: Mapped[int] = mapped_column(
        ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value: Mapped[str64] = mapped_column(primary_key=True)
    __table_args__ = (
        Index("JobTypesTaskIndex", "TQId"),
        Index("JobTypesValueIndex", "Value"),
    )


class TagsQueue(TaskQueueDBBase):
    __tablename__ = "tq_TQToTags"
    TQId: Mapped[int] = mapped_column(
        ForeignKey("tq_TaskQueues.TQId", ondelete="CASCADE"), primary_key=True
    )
    Value: Mapped[str64] = mapped_column(primary_key=True)
    __table_args__ = (
        Index("TagsTaskIndex", "TQId"),
        Index("TagsValueIndex", "Value"),
    )



from __future__ import annotations

from datetime import datetime
from zoneinfo import ZoneInfo

import sqlalchemy.types as types
from sqlalchemy import DateTime, Enum
from sqlalchemy.orm import mapped_column
from typing_extensions import Annotated

from .functions import utcnow

# Module-level constants for default timezone values
_DEFAULT_UTC = ZoneInfo("UTC")

datetime_now = Annotated[
    datetime, mapped_column(DateTime(timezone=True), server_default=utcnow())
]

str32 = Annotated[str, 32]
str64 = Annotated[str, 64]
str128 = Annotated[str, 128]
str255 = Annotated[str, 255]
str512 = Annotated[str, 512]
str1024 = Annotated[str, 1024]


def enum_column(name, enum_type, **kwargs):
    return mapped_column(name, Enum(enum_type, native_enum=False, length=16), **kwargs)


class EnumBackedBool(types.TypeDecorator):
    """Maps a ``EnumBackedBool()`` column to True/False in Python."""

    impl = types.Enum("True", "False", name="enum_backed_bool")
    cache_ok = True

    def process_bind_param(self, value, dialect) -> str:
        if value is True:
            return "True"
        elif value is False:
            return "False"
        else:
            raise NotImplementedError(value, dialect)

    def process_result_value(self, value, dialect) -> bool:
        if value == "True":
            return True
        elif value == "False":
            return False
        else:
            raise NotImplementedError(f"Unknown {value=}")


class SmarterDateTime(types.TypeDecorator):
    """A DateTime type that also accepts ISO8601 strings.

    Takes into account converting timezone aware datetime objects into
    naive form and back when needed.

    """

    impl = DateTime()
    cache_ok = True

    def __init__(
        self,
        stored_tz: ZoneInfo | None = None,
        returned_tz: ZoneInfo | None = None,
        stored_naive_sqlite=True,
        stored_naive_mysql=True,
        stored_naive_postgres=False,  # Forces timezone-awareness
    ):
        if stored_tz is None:
            stored_tz = _DEFAULT_UTC
        if returned_tz is None:
            returned_tz = _DEFAULT_UTC
        self._stored_naive_dialect = {
            "sqlite": stored_naive_sqlite,
            "mysql": stored_naive_mysql,
            "postgres": stored_naive_postgres,
        }
        self._stored_tz: ZoneInfo | None = stored_tz  # None = Local timezone
        self._returned_tz: ZoneInfo = returned_tz

    def _stored_naive(self, dialect):
        if dialect.name not in self._stored_naive_dialect:
            raise NotImplementedError(dialect.name)
        return self._stored_naive_dialect.get(dialect.name)

    def process_bind_param(self, value, dialect):
        if value is None:
            return None

        if isinstance(value, str):
            try:
                value: datetime = datetime.fromisoformat(value)
            except ValueError as err:
                raise ValueError(f"Unable to parse datetime string: {value}") from err

        if not isinstance(value, datetime):
            raise ValueError(f"Expected datetime or ISO8601 string, but got {value!r}")

        if not value.tzinfo:
            raise ValueError(
                f"Provided timestamp {value=} has no tzinfo -"
                " this is problematic and may cause inconsistencies in stored timestamps.\n"
                " Please always work with tz-aware datetimes / attach tzinfo to your datetime objects:"
                " e.g. datetime.now(tz=timezone.utc) or use datetime_obj.astimezone() with no arguments if you need to "
                "attach the local timezone to a local naive timestamp."
            )

        # Check that we need to convert the timezone to match self._stored_tz timezone:
        if self._stored_naive(dialect):
            # if self._stored_tz is None, we use our local/system timezone.
            stored_tz = self._stored_tz

            # astimezone converts to the stored timezone (local timezone if None)
            # replace strips the TZ info --> naive datetime object
            value = value.astimezone(tz=stored_tz).replace(tzinfo=None)

        return value

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        if not isinstance(value, datetime):
            raise NotImplementedError(f"{value=} not a datetime object")

        if self._stored_naive(dialect):
            # Here we add back the tzinfo to the naive timestamp
            # from the DB to make it aware again.
            if value.tzinfo is None:
                # we are definitely given a naive timestamp, so handle it.
                # add back the timezone info if stored_tz is set
                if self._stored_tz:
                    value = value.replace(tzinfo=self._stored_tz)
                else:
                    # if stored as a local time, add back the system timezone info...
                    value = value.astimezone()
            else:
                raise ValueError(
                    f"stored_naive is True for {dialect.name=}, but the database engine returned "
                    "a tz-aware datetime. You need to check the SQLAlchemy model is consistent with the DB schema."
                )

        # finally, convert the datetime according to the "returned_tz"
        value = value.astimezone(self._returned_tz)

        # phew...
        return value



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from uuid import UUID as StdUUID  # noqa: N811

import freezegun
import pytest
from uuid_utils import UUID, uuid7

from diracx.db.sql.utils import uuid7_from_datetime, uuid7_to_datetime


def frozen_uuid7() -> UUID:
    """Create a UUID7 in a way which respects the freezegun context."""
    timestamp = datetime.now(tz=timezone.utc).timestamp()
    return uuid7(int(timestamp), int((timestamp % 1) * 1e9))


class TestDatetimeToUuid7:
    """Test cases for the datetime_to_uuid7 function."""

    def test_datetime_to_uuid7_random(self):
        """Test that the datetime_to_uuid7 function returns a UUID7 with the current timestamp."""
        dt = datetime.fromisoformat("2024-01-15T12:30:45.123456+00:00")
        result = uuid7_from_datetime(dt, randomize=True)
        assert str(result)[:15] == "018d0d1a-5183-7"
        assert len(set(str(result)[15:])) > 4

    def test_datetime_to_uuid7_deterministic(self):
        """Test that the datetime_to_uuid7 function returns a UUID7 with the current timestamp."""
        dt = datetime.fromisoformat("2024-01-15T12:30:45.123456+00:00")
        result = uuid7_from_datetime(dt, randomize=False)
        assert str(result) == "018d0d1a-5183-7000-8000-000000000000"


class TestUuid7ToDatetime:
    """Test cases for the uuid7_to_datetime function."""

    def test_uuid7_now(self):
        """Test that the uuid7 function returns a UUID7 with the current timestamp."""
        test_uuid = frozen_uuid7()
        result = uuid7_to_datetime(test_uuid)
        assert result.tzinfo == timezone.utc
        assert result - datetime.now(tz=timezone.utc) < timedelta(milliseconds=2)

    @freezegun.freeze_time("2024-01-15 12:30:45.123456")
    def test_uuid7_to_datetime_with_uuid_utils_uuid(self):
        """Test conversion with uuid_utils.UUID object."""
        test_uuid = frozen_uuid7()
        result = uuid7_to_datetime(test_uuid)

        assert isinstance(result, datetime)
        assert result.tzinfo == timezone.utc
        assert result.isoformat() == "2024-01-15T12:30:45.123000+00:00"

    @freezegun.freeze_time("2024-01-15 12:30:45.123456")
    def test_uuid7_to_datetime_with_stdlib_uuid(self):
        """Test conversion with standard library UUID object."""
        test_uuid = frozen_uuid7()
        stdlib_uuid = StdUUID(str(test_uuid))

        result = uuid7_to_datetime(stdlib_uuid)
        assert isinstance(result, datetime)
        assert result.tzinfo == timezone.utc

    @freezegun.freeze_time("2024-01-15 12:30:45.123456")
    def test_uuid7_to_datetime_with_string_uuid(self):
        """Test conversion with string UUID that gets converted to uuid_utils.UUID."""
        test_uuid = frozen_uuid7()
        uuid_string = str(test_uuid)

        result = uuid7_to_datetime(uuid_string)
        assert isinstance(result, datetime)
        assert result.tzinfo == timezone.utc

    @freezegun.freeze_time("2024-01-15 12:30:45.123456")
    def test_uuid7_to_datetime_precision(self):
        """Test that the conversion maintains reasonable precision."""
        test_uuid = frozen_uuid7()
        result = uuid7_to_datetime(test_uuid)

        assert isinstance(result, datetime)
        assert result.tzinfo == timezone.utc
        assert result.isoformat() == "2024-01-15T12:30:45.123000+00:00"

    @freezegun.freeze_time("1970-01-01 00:00:00.000000")
    def test_uuid7_to_datetime_epoch_boundary(self):
        """Test conversion with UUID7 at epoch boundary."""
        test_uuid = frozen_uuid7()
        result = uuid7_to_datetime(test_uuid)

        assert isinstance(result, datetime)
        assert result.tzinfo == timezone.utc
        assert result.isoformat() == "1970-01-01T00:00:00+00:00"

    @freezegun.freeze_time("2100-01-01 00:00:00.000000")
    def test_uuid7_to_datetime_future_timestamp(self):
        """Test conversion with future timestamp."""
        test_uuid = frozen_uuid7()
        result = uuid7_to_datetime(test_uuid)

        assert isinstance(result, datetime)
        assert result.tzinfo == timezone.utc
        assert result.isoformat() == "2100-01-01T00:00:00+00:00"

    @freezegun.freeze_time("2024-01-15 12:30:45.123456")
    def test_uuid7_to_datetime_consistency(self):
        """Test that multiple calls with the same UUID return the same result."""
        test_uuid = frozen_uuid7()

        result1 = uuid7_to_datetime(test_uuid)
        result2 = uuid7_to_datetime(test_uuid)

        assert result1.isoformat() == result2.isoformat()
        assert result1 == result2

    @freezegun.freeze_time("2024-01-15 12:30:45.123456")
    def test_uuid7_to_datetime_different_input_types_same_result(self):
        """Test that different input types representing the same UUID return the same result."""
        test_uuid = frozen_uuid7()
        uuid_string = str(test_uuid)
        stdlib_uuid = StdUUID(uuid_string)

        result_from_uuid_utils = uuid7_to_datetime(test_uuid)
        result_from_string = uuid7_to_datetime(uuid_string)
        result_from_stdlib = uuid7_to_datetime(stdlib_uuid)

        # All results should be identical
        assert result_from_uuid_utils == result_from_string == result_from_stdlib

    def test_uuid7_to_datetime_invalid_input_type(self):
        """Test that invalid input types raise appropriate errors."""
        with pytest.raises(TypeError):
            uuid7_to_datetime(123)

        with pytest.raises(TypeError):
            uuid7_to_datetime(123.45)

        with pytest.raises(TypeError):
            uuid7_to_datetime([])

        with pytest.raises(TypeError):
            uuid7_to_datetime({})

    def test_uuid7_to_datetime_invalid_uuid_string(self):
        """Test that invalid UUID strings raise appropriate errors."""
        with pytest.raises(ValueError):
            uuid7_to_datetime("not-a-uuid")

        with pytest.raises(ValueError):
            uuid7_to_datetime("12345")

        with pytest.raises(ValueError):
            uuid7_to_datetime("")

    def test_uuid7_to_datetime_non_uuid7_uuid(self):
        """Test behavior with non-UUID7 UUIDs."""
        uuid4 = StdUUID("550e8400-e29b-41d4-a716-446655440000")

        with pytest.raises(ValueError, match="is not a UUIDv7"):
            uuid7_to_datetime(uuid4)

    @freezegun.freeze_time("2024-01-15 12:30:45.123456")
    def test_uuid7_to_datetime_timezone_awareness(self):
        """Test that the returned datetime is always timezone-aware and in UTC."""
        test_uuid = frozen_uuid7()
        result = uuid7_to_datetime(test_uuid)

        assert result.tzinfo is not None

        assert result.tzinfo == timezone.utc

        assert result.tzinfo.utcoffset(result) is not None

    @freezegun.freeze_time("2024-01-15 12:30:45.123456")
    def test_uuid7_to_datetime_millisecond_precision(self):
        """Test that the conversion handles millisecond precision correctly."""
        uuids = [frozen_uuid7() for _ in range(5)]
        results = [uuid7_to_datetime(uuid) for uuid in uuids]

        for result in results:
            assert isinstance(result, datetime)
            assert result.tzinfo == timezone.utc

        timestamps = [r.timestamp() for r in results]
        assert timestamps == sorted(timestamps)

    @freezegun.freeze_time("2024-01-15 12:30:45.123456")
    def test_uuid7_to_datetime_timestamp_property(self):
        """Test that the function correctly uses the timestamp property."""
        test_uuid = frozen_uuid7()
        result = uuid7_to_datetime(test_uuid)

        expected_timestamp = test_uuid.timestamp / 1000.0
        actual_timestamp = result.timestamp()

        assert abs(actual_timestamp - expected_timestamp) < 0.001

    @freezegun.freeze_time("2024-01-15 12:30:45.123456")
    def test_uuid7_to_datetime_deterministic_with_freeze(self):
        """Test that the function works deterministically with freezegun."""
        with freezegun.freeze_time("2024-01-15 12:30:45.123000"):
            uuid1 = frozen_uuid7()
            result1 = uuid7_to_datetime(uuid1)

        with freezegun.freeze_time("2024-01-15 12:30:45.123000"):
            uuid2 = frozen_uuid7()
            result2 = uuid7_to_datetime(uuid2)

        time_diff = abs((result1 - result2).total_seconds())
        assert time_diff < 1.0

    @freezegun.freeze_time("1970-01-01 00:00:00")
    def test_uuid7_to_datetime_edge_case_zero_timestamp(self):
        """Test edge case with very small timestamp values."""
        test_uuid = frozen_uuid7()
        result = uuid7_to_datetime(test_uuid)

        assert isinstance(result, datetime)
        assert result.tzinfo == timezone.utc
        assert result.isoformat() == "1970-01-01T00:00:00+00:00"



from __future__ import annotations

from collections import defaultdict
from datetime import datetime, timezone
from typing import Iterable

from sqlalchemy import delete, func, insert, select

from diracx.core.models.job import JobLoggingRecord, JobStatusReturn

from ..utils import BaseSQLDB
from .schema import JobLoggingDBBase, LoggingInfo


class JobLoggingDB(BaseSQLDB):
    """Frontend for the JobLoggingDB. Provides the ability to store changes with timestamps."""

    metadata = JobLoggingDBBase.metadata

    async def insert_records(
        self,
        records: list[JobLoggingRecord],
    ):
        """Bulk insert entries to the JobLoggingDB table."""
        # First, fetch the maximum SeqNums for the given job_ids
        seqnum_stmt = (
            select(
                LoggingInfo.job_id, func.coalesce(func.max(LoggingInfo.seq_num) + 1, 1)
            )
            .where(LoggingInfo.job_id.in_([record.job_id for record in records]))
            .group_by(LoggingInfo.job_id)
        )

        seqnums = {
            jid: seqnum for jid, seqnum in (await self.conn.execute(seqnum_stmt))
        }
        # IF a seqnum is not found, then assume it does not exist and the first sequence number is 1.
        # https://docs.sqlalchemy.org/en/20/orm/queryguide/dml.html#orm-bulk-insert-statements
        values = []
        for record in records:
            if record.job_id not in seqnums:
                seqnums[record.job_id] = 1

            values.append(
                {
                    "JobID": record.job_id,
                    "SeqNum": seqnums[record.job_id],
                    "Status": record.status,
                    "MinorStatus": record.minor_status,
                    "ApplicationStatus": record.application_status[:255],
                    "StatusTime": record.date,
                    "StatusTimeOrder": record.date,
                    "StatusSource": record.source[:32],
                }
            )
            seqnums[record.job_id] = seqnums[record.job_id] + 1

        await self.conn.execute(
            insert(LoggingInfo),
            values,
        )

    async def get_records(self, job_ids: list[int]) -> dict[int, JobStatusReturn]:
        """Return a Status,MinorStatus,ApplicationStatus,StatusTime,Source tuple.

        For each record found for job specified by its jobID in historical order.
        """
        # We could potentially use a group_by here, but we need to post-process the
        # results later.
        stmt = (
            select(
                LoggingInfo.job_id,
                LoggingInfo.status,
                LoggingInfo.minor_status,
                LoggingInfo.application_status,
                LoggingInfo.status_time,
                LoggingInfo.source,
            )
            .where(LoggingInfo.job_id.in_(job_ids))
            .order_by(LoggingInfo.status_time_order, LoggingInfo.status_time)
        )
        rows = await self.conn.execute(stmt)

        values = defaultdict(list)
        for (
            job_id,
            status,
            minor_status,
            application_status,
            status_time,
            status_source,
        ) in rows:
            values[job_id].append(
                [
                    status,
                    minor_status,
                    application_status,
                    status_time.replace(tzinfo=timezone.utc),
                    status_source,
                ]
            )

        # If no value has been set for the application status in the first place,
        # We put this status to unknown
        res: dict = defaultdict(list)
        for job_id, history in values.items():
            if history[0][2] == "idem":
                history[0][2] = "Unknown"

            # We replace "idem" values by the value previously stated
            for i in range(1, len(history)):
                for j in range(3):
                    if history[i][j] == "idem":
                        history[i][j] = history[i - 1][j]

            # And we replace arrays with tuples
            for (
                status,
                minor_status,
                application_status,
                status_time,
                status_source,
            ) in history:
                res[job_id].append(
                    JobStatusReturn(
                        Status=status,
                        MinorStatus=minor_status,
                        ApplicationStatus=application_status,
                        StatusTime=status_time,
                        Source=status_source,
                    )
                )

        return res

    async def delete_records(self, job_ids: list[int]):
        """Delete logging records for given jobs."""
        stmt = delete(LoggingInfo).where(LoggingInfo.job_id.in_(job_ids))
        await self.conn.execute(stmt)

    async def get_wms_time_stamps(
        self, job_ids: Iterable[int]
    ) -> dict[int, dict[str, datetime]]:
        """Get TimeStamps for job MajorState transitions for multiple jobs at once.

        return a {JobID: {State:timestamp}} dictionary.
        """
        result: defaultdict[int, dict[str, datetime]] = defaultdict(dict)
        stmt = select(
            LoggingInfo.job_id, LoggingInfo.status, LoggingInfo.status_time_order
        ).where(LoggingInfo.job_id.in_(job_ids))
        for job_id, event, etime in await self.conn.execute(stmt):
            result[job_id][event] = etime
        return dict(result)
